/*
Our project, End-to-end workout, is a smart fitness tracking system that monitors
workout repetitions using an IMU sensor.
By detecting direction changes, it counts reps for exercises
like push-ups and sit-ups. We've integrated Blynk to display real-time workout
progress and control various features remotely.
The system also connects to a Control4 automation system,
allowing users to activate devices such as water tanks
and lights during their workout.
Workout data is sent to a Telegram bot via an HTTP request,
providing users with an automated summary of their progress.

Features:
Real-time workout tracking with rep counting for push-ups and sit-ups
Blynk integration for monitoring and control
Control4 automation to activate connected devices
Workout summary sent to a Telegram bot

Additional Integrations: Can toggle a water tank and shower light using Control4 smart home automation.

Circuit:
  * Accelerometer (IMU): SCL (IO22), SDA (IO21)
  * Speaker: IO26
  * LED Strip: IO16
  * Blue LED: IO15

  Video link: https://drive.google.com/file/d/1yj5HabeXo-aIdZIvFVQdgoSLfnLDi5GO/view?usp=sharing
  Created By:
        Lior Zoaretz
        Danielle Joy Levy
*/

// Blynk configuration
#define BLYNK_TEMPLATE_ID "your_blynk_template_id"
#define BLYNK_TEMPLATE_NAME "End to End Workout"
#define BLYNK_AUTH_TOKEN "your_blynk_auth_token"

#include "pitches.h"
#include <Arduino.h>
#include <Adafruit_NeoPixel.h>
#include <Adafruit_Sensor.h>
#include <HTTPClient.h>
#include <WiFi.h>
#include <WiFiClient.h>
#include <WiFiClientSecure.h>
#include <BlynkSimpleEsp32.h>
#include "I2Cdev.h"
#include "MPU6050.h"

// wifi configuration
// fill your wifi details
char ssid[] = "your_wifi_ssid";
char pass[] = "your_wifi_password";

// control4 configuration
// fill your control4 details
const char *control4LocalUrl = "https://your_wifi_ip_address/api/v1/items/";
const char *bearerToken = "your_control4_bearer_token"; // add your control4 bearer token

// control4 properties ids
const char *waterTankId = "219";
const char *lightInShowerId = "229";

// Make.com configuration
// fill your Make.com scenario webhook
const char *makeUrl = "https://hook.eu2.make.com/your_make_scenario_webhook?";
const char *pushUpsSuffix = "pushUps=";
const char *sitUpsSuffix = "&sitUps=";
const char *setsSuffix = "&sets=";

// configuration of the PIN
#define SPEAKER_PIN 26 // speaker pin
#define SIGNAL_LED_PIN 15 // blue LED
#define LED_STRIP_PIN 16 // led strip pin
#define LED_COUNT 13 // number of leds in the strip

// speaker configuration
const int SPEAKER_PWM_CHANNEL = 0;
int successMelodyNotes[] = {
    NOTE_E7, NOTE_G7, NOTE_E7, NOTE_C7, NOTE_D7, NOTE_G6,
    NOTE_C7, NOTE_E6, NOTE_A6, NOTE_B6, NOTE_A6, NOTE_GS6, NOTE_B6, NOTE_GS6
};

// configure Adafruit led strip
Adafruit_NeoPixel strip(LED_COUNT, LED_STRIP_PIN, NEO_GRB + NEO_KHZ800);

// IMU configuration
#define OUTPUT_READABLE_ACCELGYRO

MPU6050 accelgyro;

// IMU axis configuration
int16_t ax, ay, az;
int16_t gx, gy, gz;

// accelerator state variables
float yAxisAccelerator = 0;
int countDirectionMovement = 0;

// workout state variables
bool isPushUpsWorkoutStarted = false;
bool isSitUpsWorkoutStarted = false;
int numberOfPushUps = 0;
int numberOfSitUps = 0;
int lastSentNumberOfPushUps = 0;
int lastSentNumberOfSitUps = 0;
int lastSetCount = 1;

// workout send time interval variables
unsigned long lastPushUpsSendTime = 0;
unsigned long lastSitUpsSendTime = 0;
unsigned long sendInterval = 500;

// Blynk functions
// handle Start sit ups workout switch widget
BLYNK_WRITE(V0)
{
  // variable of pin V0 value
  int pinValue = param.asInt();
  // if switch widget is off reset Number of sit ups to 0
  if (pinValue == 0)
  {
    isSitUpsWorkoutStarted = false;
    digitalWrite(SIGNAL_LED_PIN, LOW); // turn the LED off
    Blynk.virtualWrite(V3, 0); // Number of sit ups pin V3
  }
  // if switch widget is on then start sit ups workout
  else
  {
    isSitUpsWorkoutStarted = true;
    digitalWrite(SIGNAL_LED_PIN, HIGH); // turn the LED on
    Blynk.virtualWrite(V1, 0); // turn off Start push ups workout
  }
  numberOfSitUps = 0;
}

// handle Start push ups workout switch widget
BLYNK_WRITE(V1)
{
  // variable of pin V1 value
  int pinValue = param.asInt();
  // if switch widget is off reset Number of push ups to 0
  if (pinValue == 0)
  {
    isPushUpsWorkoutStarted = false;
    digitalWrite(SIGNAL_LED_PIN, LOW); // turn the LED off
    Blynk.virtualWrite(V2, 0); // Number of sit ups pin V2
  }
  // if switch widget is on then start push ups workout
  else
  {
    isPushUpsWorkoutStarted = true;
    digitalWrite(SIGNAL_LED_PIN, HIGH); // turn the LED on
    Blynk.virtualWrite(V0, 0); // turn off Start sit ups workout
  }
  numberOfPushUps = 0;
}

// handles Reset switch widget
BLYNK_WRITE(V5)
{
  // variable of pin V5 value
  int command = param.asInt();
  // if the Reset switch is on then end the workout
  if (command == 1)
  {
    // turn on water tank and the light in the shower at the end of the workout via control4
    turnOnWaterTank();
    turnOnLightInShower();
    // send final set (of workout) summary to telegram bot
    sendWorkoutDataToTelegramBot(numberOfPushUps, numberOfSitUps, lastSetCount);
    Blynk.virtualWrite(V4, 1); // reset set count
    Blynk.virtualWrite(V5, 0); // turn off reset switch
    Blynk.virtualWrite(V0, 0); // turn off start sit-ups switch
    Blynk.virtualWrite(V1, 0); // turn off start push-ups switch
    Blynk.virtualWrite(V3, 0); // reset sit-ups count
    Blynk.virtualWrite(V2, 0); // Reset push-ups count
    // reset workout state variables
    isPushUpsWorkoutStarted = false;
    isSitUpsWorkoutStarted = false;
    // reset workout set state variable
    lastSetCount = 1;
    digitalWrite(SIGNAL_LED_PIN, LOW); // turn the LED off
  }
}

// handles Set switch widget to start a new set
BLYNK_WRITE(V6)
{
  // variable of pin V6 value
  int command = param.asInt();
  // if the Set switch is on then start a new set in the workout
  if (command == 1)
  {
    // send set summary to telegram bot
    sendWorkoutDataToTelegramBot(numberOfPushUps, numberOfSitUps, lastSetCount);
    lastSetCount++; // update the set count in the workout
    Blynk.virtualWrite(V4, lastSetCount); // add set to the number input widget in Blynk
    // reset set state variables
    numberOfPushUps = 0;
    numberOfSitUps = 0;
    resetPins();
    Blynk.virtualWrite(V6, 0); // turn off reset switch
  }
}

void setup()
{
  // IMU initialization
  Serial.begin(115200);
  Serial.println("Initializing IMU");
  accelgyro.initialize();

  // initialize speaker pin
  ledcAttachPin(SPEAKER_PIN, SPEAKER_PWM_CHANNEL);

  // initialize signal led pin as an output
  pinMode(SIGNAL_LED_PIN, OUTPUT);

  // connect to wifi
  connectToWifi();

  // connect to Blynk
  Blynk.begin(BLYNK_AUTH_TOKEN, ssid, pass);

  // configure all pins to initial state
  Blynk.virtualWrite(V0, 0); // reset Start sit ups in Virtual Pin V0
  Blynk.virtualWrite(V1, 0); // reset Start push ups in Virtual Pin V1
  Blynk.virtualWrite(V2, 0); // reset push-ups count in Virtual Pin V2
  Blynk.virtualWrite(V3, 0); // reset sit-ups count in Virtual Pin V3
  Blynk.virtualWrite(V4, 1); // reset set number to 1 in Virtual Pin V4
  Blynk.setProperty(V5, "isHidden", true); // hide Reset switch in the Blynk dashboard
  Blynk.setProperty(V6, "isHidden", true); // hide Start new set switch in the Blynk dashboard
}

void loop()
{
  Blynk.run();
  recordPushUps();
  recordSitUps();
}

void connectToWifi()
{
  // connecting to wifi
  WiFi.mode(WIFI_STA);
  WiFi.begin(ssid, pass);
  Serial.print("Connecting to wifi...");
  while (WiFi.status() != WL_CONNECTED)
  {
    Serial.print(".");
    delay(1000);
  }
  Serial.println("successfully connected to wifi");
}

// record number of sit-ups in a single set
void recordSitUps()
{
    // send data to Blynk at specified intervals
    if (millis() - lastSitUpsSendTime > sendInterval && isSitUpsWorkoutStarted)
    {
        accelgyro.getMotion6(&ax, &ay, &az, &gx, &gy, &gz);

        // detect if any movement recognized in y axis
        if (ay < 0 && yAxisAccelerator > 0)
        {
            countDirectionMovement++;
        }
        // detect if any movement recognized in y axis
        if (yAxisAccelerator < 0 && ay > 0) {
            countDirectionMovement++;
        }

        // if there were 3 direction changes it is a sit-up
        if (countDirectionMovement >= 3)
        {
            numberOfSitUps++;
            countDirectionMovement = 0;
        }

        // in order to make sure blynk is not spammed
        if (numberOfSitUps != lastSentNumberOfSitUps)
        {
            Blynk.virtualWrite(V3, numberOfSitUps); // send the number of sit-ups to blynk Virutal Pin called V3
            lastSentNumberOfSitUps = numberOfSitUps;
            triggerSuccessSequence();
        }
        yAxisAccelerator = ay;
        lastSitUpsSendTime = millis();
    }
}

// record number of push-ups in a single set
void recordPushUps()
{
    // send data to Blynk at specified intervals
    if (millis() - lastPushUpsSendTime > sendInterval && isPushUpsWorkoutStarted)
    {
        accelgyro.getMotion6(&ax, &ay, &az, &gx, &gy, &gz);

        // detect if any movement recognized in y axis
        if (ay < 0 && yAxisAccelerator > 0)
        {
            countDirectionMovement++;
        }
        // detect if any movement recognized in x axis
        if (yAxisAccelerator < 0 && ay > 0) {
            countDirectionMovement++;
        }

        // if there were 2 direction changes it is a push-up
        if (countDirectionMovement >= 2)
        {
            numberOfPushUps++;
            countDirectionMovement = 0;
        }

        // in order to make sure blynk is not spammed
        if (numberOfPushUps != lastSentNumberOfPushUps)
        {
            Blynk.virtualWrite(V2, numberOfPushUps); // send the number of push-ups to blynk Virutal Pin called V2
            lastSentNumberOfPushUps = numberOfPushUps;
            triggerSuccessSequence();
        }
        yAxisAccelerator = ay;
        lastPushUpsSendTime = millis();
    }
}

// reset set state number input widgets values
void resetPins()
{
  if (isPushUpsWorkoutStarted)
  {
    Blynk.virtualWrite(V2, 0); // reset push-ups count
  }
  else
  {
    Blynk.virtualWrite(V3, 0); // reset sit-ups count
  }
}

// turn on water tank via control4
void turnOnWaterTank()
{
  triggerControl4OnOff(waterTankId, 1);
}

// turn on light in shower via control4
void turnOnLightInShower()
{
  triggerControl4OnOff(lightInShowerId, 1);
}

// send post request to set control4 item state
void triggerControl4OnOff(const char *itemId, int isON)
{
  WiFiClientSecure *client = new WiFiClientSecure;
  if (client)
  {
    client->setInsecure();
    {
      HTTPClient https;
      if (https.begin(*client, String(control4LocalUrl) + itemId + "/commands"))
      {
        // adding headers
        https.addHeader("Content-Type", "application/json"); // Set JSON header
        https.addHeader("Authorization", "Bearer " + String(bearerToken));
        String postPayload = "";
        // edit the request payload according to isOn variable
        if (isON)
        {
          String postPayload = "{\"async\": true, \"command\": \"ON\", \"tParams\": {}}";
        }
        else
        {
          String postPayload = "{\"async\": true, \"command\": \"OFF\", \"tParams\": {}}";
        }
        // sending post request
        int httpCode = https.POST(postPayload);
        // handle post response
        if (httpCode > 0)
        {
          Serial.printf("POST response code: %d\n", httpCode);
          if (httpCode == HTTP_CODE_OK || httpCode == HTTP_CODE_CREATED)
          {
            String response = https.getString();
            Serial.println("Response: " + response);
          }
        }
        // post request failed
        else
        {
          Serial.printf("POST request failed, error: %s\n", https.errorToString(httpCode).c_str());
        }
        https.end();
      }
      else
      {
        Serial.println("Unable to connect to control4");
      }
    }
    delete client;
  }
  else
  {
    Serial.println("Unable to create https client");
  }
}

// send a single set summary to telegram bot
void sendWorkoutDataToTelegramBot(int pushUpsNum, int sitUpsNum, int sets)
{
  WiFiClientSecure *client = new WiFiClientSecure;
  if (client)
  {
    client->setInsecure();
    {
      HTTPClient https;
      if (https.begin(*client, String(makeUrl) + String(pushUpsSuffix) + pushUpsNum + String(sitUpsSuffix) + sitUpsNum + String(setsSuffix) + sets))
      {
        // send get request
        int httpCode = https.GET();
        // handle get response
        if (httpCode > 0)
        {
          Serial.printf("GET response code: %d\n", httpCode);
          if (httpCode == HTTP_CODE_OK || httpCode == HTTP_CODE_MOVED_PERMANENTLY)
          {
            String payload = https.getString();
            Serial.println(payload);
          }
        }
        // get request failed
        else
        {
          Serial.printf("GET request failed, error: %s\n", https.errorToString(httpCode).c_str());
        }
        https.end();
      }
      else
      {
        Serial.println("Unable to connect to make.com");
      }
    }
    delete client;
  }
  else
  {
    Serial.println("Unable to create https client");
  }
}

// plays success lights and sound
void triggerSuccessSequence()
{
    startNotesMelody(successMelodyNotes, 14); // play success melody
    flashSuccessLights(); // flash success lights
}

// flash the green color when the trainee succeeded
void flashSuccessLights()
{
    for (byte flashIndex = 0; flashIndex < 3; flashIndex++)
    {
        for (byte ledIndex = 0; ledIndex < LED_COUNT; ledIndex++)
        {
            strip.setPixelColor(ledIndex, strip.Color(0, 255, 0)); // green light
        }
        strip.show();
        delay(150);
        turnOffLeds();
        delay(150);
    }
}

// turns off the leds in the led strip
void turnOffLeds()
{
    // turn off every led in the led strip
    for (int ledsIndex = 0; ledsIndex < LED_COUNT; ledsIndex++)
    {
        strip.setPixelColor(ledsIndex, strip.Color(0, 0, 0));
    }
    strip.show();
}

// plays every note in the the given melody using pwm channel
void startNotesMelody(int *melody, int melodyLen)
{
    for (int note = 0; note < melodyLen; note++)
    {
        int noteDuration = 50;
        int notesDelay = noteDuration * 0.20;
        ledcWriteTone(SPEAKER_PWM_CHANNEL, melody[note]);
        delay(noteDuration);
        ledcWrite(SPEAKER_PWM_CHANNEL, 0);
        delay(notesDelay);
    }
}
